@@ -3,7 +3,7 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -529,6 +529,11 @@ int check_account(account_t *acc, int sendmail_mode, char **errstr)
 	*errstr = xasprintf(_("host not set"));
 	return CONF_ESYNTAX;
     }
+    if (acc->port == 0)
+    {
+	*errstr = xasprintf(_("port not set"));
+	return CONF_ESYNTAX;
+    }
     if (sendmail_mode && !acc->from)
     {
 	*errstr = xasprintf(_("envelope-from address is missing"));
@@ -1043,6 +1048,7 @@ int read_conffile(const char *conffile, FILE *f, list_t **acc_list,
 	    }
 	    else
 	    {
+		acc->auth_mech = NULL;
 		*errstr = xasprintf(
 	    		_("line %d: invalid argument %s for command %s"),
     			line, arg, cmd);
