@@ -3,7 +3,7 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -39,17 +39,13 @@ extern int optind;
 #ifdef ENABLE_NLS
 #include <locale.h>
 #endif
-#include <sysexits.h>
-#ifndef EX_OK
-#define EX_OK 0
-#endif
 #ifdef HAVE_SYSLOG
 #include <syslog.h>
 #endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#ifdef __MINGW32__
+#ifdef W32_NATIVE
 #include <io.h>
 #include <fcntl.h>
 #include <windows.h>
@@ -62,6 +58,7 @@ extern int optind;
 #include <netdb.h>
 #include <arpa/inet.h>
 #endif
+#include <sysexits.h>
 
 #include "getpass.h"
 #include "gettext.h"
@@ -79,7 +76,7 @@ extern int optind;
 #endif /* HAVE_SSL */
 
 /* Default file names. */
-#ifdef __MINGW32__
+#ifdef W32_NATIVE
 #define SYSCONFFILE	"msmtprc.txt"
 #define USERCONFFILE	"msmtprc.txt"
 #define NETRCFILE	"netrc.txt"
@@ -453,6 +450,13 @@ int msmtp_rmqs(account_t *acc, int debug, const char *rmqs_argument,
     /* create a new smtp_server_t */
     srv = smtp_new(debug ? stdout : NULL, acc->protocol);
 
+    /* connect */
+    if ((e = smtp_connect(&srv, acc->host, acc->port, acc->timeout,
+		    NULL, NULL, errstr)) != NET_EOK)
+    {
+	return exitcode_net(e);
+    }
+
     /* prepare tls */
 #ifdef HAVE_SSL
     if (acc->tls)
@@ -465,13 +469,6 @@ int msmtp_rmqs(account_t *acc, int debug, const char *rmqs_argument,
     }
 #endif /* HAVE_SSL */
 
-    /* connect */
-    if ((e = smtp_connect(&srv, acc->host, acc->port, acc->timeout,
-		    NULL, NULL, errstr)) != NET_EOK)
-    {
-	return exitcode_net(e);
-    }
-
     /* start tls for ssmtp servers */
 #ifdef HAVE_SSL
     if (acc->tls && acc->tls_nostarttls)
@@ -623,6 +620,15 @@ int msmtp_serverinfo(account_t *acc, int debug, list_t **msg, char **errstr)
     /* create a new smtp_server_t */
     srv = smtp_new(debug ? stdout : NULL, acc->protocol);
 
+    /* connect */
+    if ((e = smtp_connect(&srv, acc->host, acc->port, acc->timeout,
+		    &server_canonical_name, &server_address, errstr))
+	    != NET_EOK)
+    {
+	e = exitcode_net(e);
+       	goto error_exit;
+    }
+
     /* prepare tls */
 #ifdef HAVE_SSL
     if (acc->tls)
@@ -637,15 +643,6 @@ int msmtp_serverinfo(account_t *acc, int debug, list_t **msg, char **errstr)
     }
 #endif /* HAVE_SSL */
 
-    /* connect */
-    if ((e = smtp_connect(&srv, acc->host, acc->port, acc->timeout,
-		    &server_canonical_name, &server_address, errstr))
-	    != NET_EOK)
-    {
-	e = exitcode_net(e);
-       	goto error_exit;
-    }
-
     /* start tls for ssmtp servers */
 #ifdef HAVE_SSL
     if (acc->tls && acc->tls_nostarttls)
@@ -1810,19 +1807,20 @@ void msmtp_log_to_file(const char *logfile, const char *loginfo)
     FILE *f;
     time_t t;
     struct tm *tm;
-    const char *failure_reason;
+    char *failure_reason;
     char time_str[64];
-    
+    int e;
     
     /* get time */
     if ((t = time(NULL)) < 0)
     {
-	failure_reason = strerror(errno);
+	failure_reason = xasprintf(_("cannot get system time: %s"), 
+		strerror(errno));
 	goto log_failure;
     }
     if (!(tm = localtime(&t)))
     {
-	failure_reason = _("cannot convert UTC time to local time");
+	failure_reason = xstrdup(_("cannot convert UTC time to local time"));
 	goto log_failure;
     }
     (void)strftime(time_str, sizeof(time_str), "%b %d %H:%M:%S", tm);
@@ -1834,22 +1832,36 @@ void msmtp_log_to_file(const char *logfile, const char *loginfo)
     }
     else
     {
-	if (!(f = fopen(logfile, "a")) 
-		|| lock_file(f, OSENV_LOCK_WRITE, 10) != 0)
+	if (!(f = fopen(logfile, "a")))
 	{
-	    failure_reason = strerror(errno);
+	    failure_reason = xasprintf(_("cannot open: %s"), strerror(errno));
+	    goto log_failure;
+	}
+	if ((e = lock_file(f, OSENV_LOCK_WRITE, 10)) != 0)
+	{
+	    if (e == 1)
+	    {
+		failure_reason = xasprintf(
+			_("cannot lock (tried for %d seconds): %s"), 
+			10, strerror(errno));
+	    }
+	    else
+	    {
+		failure_reason = xasprintf(_("cannot lock: %s"), 
+			strerror(errno));
+	    }
 	    goto log_failure;
 	}
     }
     if ((fputs(time_str, f) == EOF) || (fputc(' ', f) == EOF)
 	|| (fputs(loginfo, f) == EOF) || (fputc('\n', f) == EOF))
     {
-	failure_reason = _("output error");
+	failure_reason = xstrdup(_("output error"));
 	goto log_failure;
     }
     if (f != stdout && fclose(f) != 0)
     {
-	failure_reason = strerror(errno);
+	failure_reason = xstrdup(strerror(errno));
 	goto log_failure;
     }
 
@@ -1858,6 +1870,7 @@ void msmtp_log_to_file(const char *logfile, const char *loginfo)
     /* error exit target */
 log_failure:
     print_error(_("cannot log to %s: %s"), logfile, failure_reason);
+    free(failure_reason);
     if (loginfo)
     {
 	print_error(_("log info was: %s"), loginfo);
@@ -1967,29 +1980,26 @@ void msmtp_print_version(void)
     char *homedir;
     char *userconffile;
     
-    printf(_("%s version %s\nNLS: "), PACKAGE_NAME, VERSION);
-#ifdef ENABLE_NLS
-    printf(_("enabled, LOCALEDIR is %s\n"), LOCALEDIR);
-#else
-    printf("disabled\n");
-#endif
-    printf(_("TLS/SSL library: %s\n"
-		"Authentication library: %s\n"
-		"Supported authentication methods:\n"),
+    printf(_("%s version %s\n"), PACKAGE_NAME, VERSION);
+    /* TLS/SSL support */
+    printf(_("TLS/SSL library: %s\n"),
 #ifdef HAVE_GNUTLS
 	    "GnuTLS"
 #elif defined (HAVE_OPENSSL)
-	    "OpenSSL"
+    	    "OpenSSL"
 #else
 	    _("none")
 #endif
-	    ,
+	  );
+    /* Authentication support */
+    printf(_("Authentication library: %s\n"
+		"Supported authentication methods:\n"),
 #ifdef USE_GSASL
-	    _("GNU SASL")
+	    "GNU SASL"
 #else
 	    _("built-in")
 #endif /* USE_GSASL */
-	      );
+	  );
     if (smtp_client_supports_authmech("PLAIN"))
     {
 	printf("plain ");
@@ -2018,6 +2028,23 @@ void msmtp_print_version(void)
     {
 	printf("ntlm ");
     }
+    printf("\n");
+    /* Internationalized Domain Names support */
+    printf(_("IDN support: "));
+#ifdef USE_LIBIDN
+    printf(_("enabled"));
+#else
+    printf(_("disabled"));
+#endif
+    printf("\n");
+    /* Native language support */
+    printf(_("NLS: "));
+#ifdef ENABLE_NLS
+    printf(_("enabled"));
+    printf(_(", LOCALEDIR is %s"), LOCALEDIR);
+#else
+    printf(_("disabled"));
+#endif
     printf("\n");
     sysconfdir = get_sysconfdir();
     sysconffile = get_filename(sysconfdir, SYSCONFFILE);
@@ -2029,7 +2056,8 @@ void msmtp_print_version(void)
     printf(_("User configuration file name: %s\n"), userconffile);
     free(userconffile);
     free(homedir);
-    printf(_("\nCopyright (C) 2004, 2005 Martin Lambers and others.\n"
+    printf("\n");
+    printf(_("Copyright (C) 2006 Martin Lambers and others.\n"
     		"This is free software; see the source for copying "
 		"conditions.  There is NO\n"
 		"warranty; not even for MERCHANTABILITY or FITNESS FOR A "
@@ -2056,66 +2084,67 @@ void msmtp_print_help(void)
 	    "  Print information about a server.\n"
 	    "Remote Message Queue Starting mode:\n"
 	    "  %s [option...] --rmqs=host|@domain|#queue\n"
-	    "  Send a Remote Message Queue Starting request to a server\n"
+	    "  Send a Remote Message Queue Starting request to a server.\n"
             "\nOPTIONS:\n\n"
 	    "General options:\n"
-	    "  --version                    print version\n"
-	    "  --help                       print help\n"
-	    "  -P, --pretend                print configuration info and exit\n"
-	    "  -d, --debug                  print debugging information\n"
+	    "  --version                    Print version.\n"
+	    "  --help                       Print help.\n"
+	    "  -P, --pretend                Print configuration info and "
+	        "exit.\n"
+	    "  -d, --debug                  Print debugging information.\n"
 	    "Changing the mode of operation:\n"
-	    "  -S, --serverinfo             print information about the "
-	    	"server\n"
-	    "  --rmqs=host|@domain|#queue   send a Remote Message Queue "
-	    	"Starting request\n"
+	    "  -S, --serverinfo             Print information about the "
+	    	"server.\n"
+	    "  --rmqs=host|@domain|#queue   Send a Remote Message Queue "
+	    	"Starting request.\n"
             "Configuration options:\n"
-	    "  -C, --file=filename          set configuration file\n"
-	    "  -a, --account=id             use the given account instead of "
+	    "  -C, --file=filename          Set configuration file.\n"
+	    "  -a, --account=id             Use the given account instead of "
 	        "the account\n"
 	    "                               named \"default\"; its settings "
 	        "may be changed\n"
-	    "                               with command line options\n"
-	    "  --host=hostname              set the server, use only command "
+	    "                               with command line options.\n"
+	    "  --host=hostname              Set the server, use only command "
 	    	"line settings;\n"
 	    "                               do not use any configuration file "
-	    	"data\n"
-	    "  --port=number                set port number\n"
-	    "  --timeout=(off|seconds)      set/unset network timeout in "
-		"seconds\n"
-	    "  --protocol=(smtp|lmtp)       use the given sub protocol\n"
-	    "  --domain=string              set the argument of EHLO or LHLO "
-	    	"command\n"
-	    "  --auth[=(on|off|method)]     enable/disable authentication and "
+	    	"data.\n"
+	    "  --port=number                Set port number.\n"
+	    "  --timeout=(off|seconds)      Set/unset network timeout in "
+		"seconds.\n"
+	    "  --protocol=(smtp|lmtp)       Use the given sub protocol.\n"
+	    "  --domain=string              Set the argument of EHLO or LHLO "
+	    	"command.\n"
+	    "  --auth[=(on|off|method)]     Enable/disable authentication and "
 	    	"optionally\n"
-	    "                               choose the method\n"
-	    "  --user=[username]            set/unset user name for "
-	    	"authentication\n"
-	    "  --tls[=(on|off)]             enable/disable TLS encryption\n"
-	    "  --tls-trust-file=[file]      set/unset trust file for TLS\n"
-            "  --tls-key-file=[file]        set/unset private key file for "
-	    	"TLS\n"
-	    "  --tls-cert-file=[file]       set/unset private cert file for "
-	    	"TLS\n"
-	    "  --tls-certcheck[=(on|off)]   enable/disable server certificate "
-	    	"checks for TLS\n"
-	    "  --tls-starttls[=(on|off)]    enable/disable STARTTLS for TLS\n"
+	    "                               choose the method.\n"
+	    "  --user=[username]            Set/unset user name for "
+	    	"authentication.\n"
+	    "  --tls[=(on|off)]             Enable/disable TLS encryption.\n"
+	    "  --tls-trust-file=[file]      Set/unset trust file for TLS.\n"
+            "  --tls-key-file=[file]        Set/unset private key file for "
+	    	"TLS.\n"
+	    "  --tls-cert-file=[file]       Set/unset private cert file for "
+	    	"TLS.\n"
+	    "  --tls-certcheck[=(on|off)]   Enable/disable server certificate "
+	    	"checks for TLS.\n"
+	    "  --tls-starttls[=(on|off)]    Enable/disable STARTTLS for TLS.\n"
             "Options specific to sendmail mode:\n"
-	    "  --auto-from[=(on|off)]       enable/disable automatic "
-	        "envelope-from addresses\n"
-	    "  -f, --from=address           set envelope from address\n"
-	    "  --maildomain=[domain]        set the domain for automatic "
+	    "  --auto-from[=(on|off)]       Enable/disable automatic "
+	        "envelope-from addresses.\n"
+	    "  -f, --from=address           Set envelope from address.\n"
+	    "  --maildomain=[domain]        Set the domain for automatic "
 	        "envelope from\n"
-	    "                               addresses\n"
-	    "  -N, --dsn-notify=(off|cond)  set/unset DSN conditions\n"
-	    "  -R, --dsn-return=(off|ret)   set/unset DSN amount\n"
-	    "  --keepbcc[=(on|off)]         enable/disable preservation of the "
-	    	"Bcc header\n"
-            "  -X, --logfile=[file]         set/unset log file\n"
-	    "  --syslog[=(on|off|facility)] enable/disable/configure syslog "
-	    	"logging\n"
-	    "  -t, --read-recipients        read additional recipients from "
-	    	"the mail\n"
-	    "  --                           end of options\n"
+	    "                               addresses.\n"
+	    "  -N, --dsn-notify=(off|cond)  Set/unset DSN conditions.\n"
+	    "  -R, --dsn-return=(off|ret)   Set/unset DSN amount.\n"
+	    "  --keepbcc[=(on|off)]         Enable/disable preservation of the "
+	    	"Bcc header.\n"
+            "  -X, --logfile=[file]         Set/unset log file.\n"
+	    "  --syslog[=(on|off|facility)] Enable/disable/configure syslog "
+	    	"logging.\n"
+	    "  -t, --read-recipients        Read additional recipients from "
+	    	"the mail.\n"
+	    "  --                           End of options.\n"
             "Accepted but ignored: -A, -B, -bm, -F, -G, -h, -i, -L, -m, -n, "
 	    	"-O, -o, -v\n"
 	    "\nReport bugs to <%s>.\n"),
@@ -2419,6 +2448,7 @@ int msmtp_cmdline(msmtp_cmdline_conf_t *conf, int argc, char *argv[])
 		}
 		else
 		{
+		    conf->cmdline_account->auth_mech = NULL;
 		    print_error(_("invalid argument %s for %s"), 
 			    optarg, "--auth");
 		    error_code = 1;
@@ -3000,7 +3030,7 @@ int main(int argc, char *argv[])
 
     
     /* Avoid the side effects of text mode interpretations on DOS systems. */
-#ifdef __MINGW32__
+#ifdef W32_NATIVE
     _setmode(_fileno(stdin), _O_BINARY);
 #elif defined DJGPP
     setmode(fileno(stdin), O_BINARY);
@@ -3132,10 +3162,13 @@ int main(int argc, char *argv[])
 		account->port = se ? ntohs(se->s_port) : 25;
 	    }
 	}
-	else /* LMTP */
+	else /* LMTP. Has no default port as of 2006-06-17. */
 	{
 	    se = getservbyname("lmtp", NULL);
-	    account->port = se ? ntohs(se->s_port) : 2003;
+	    if (se)
+	    {
+		account->port = ntohs(se->s_port);
+	    }
 	}
     }
     if (conf.sendmail && account->auto_from)
