@@ -3,7 +3,7 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -17,11 +17,8 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with this program; if not, write to the Free Software
- *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- *
- *   msmtp is released under the GPL with the additional exemption that
- *   compiling, linking, and/or using OpenSSL is allowed.
+ *   along with this program; if not, write to the Free Software Foundation,
+ *   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -42,10 +39,12 @@ extern int errno;
 #include <gsasl.h>
 #else
 #include "base64.h"
-#include "crypto.h"
+#include "hmac.h"
 #endif
 
+#include "gettext.h"
 #include "xalloc.h"
+#include "xvasprintf.h"
 
 #include "list.h"
 #include "net.h"
@@ -74,8 +73,8 @@ extern int errno;
 
 /* This defines the length of the *output* buffer for SMTP commands, without
  * CRLF. According to RFC 2821, SMTP commands may contain at most 512
- * characters including CRLF, thus the *maximum* size should be 510 characters.
- * Bus this is not sufficient for some authentication mechanisms, notably
+ * characters including CRLF, thus the maximum size should be 510 characters.
+ * But this is not sufficient for some authentication mechanisms, notably
  * GSSAPI, so this value must be higher.
  */
 #define SMTP_MAXCMDLEN 1022
@@ -93,7 +92,7 @@ extern int errno;
  * characters + CRLF. Plus one character for '\0' = 1001 characters.
  * All lines should fit in a buffer of this size.
  * However, this length is not a limit; smtp_send_mail() will accept 
- * arbitrary long lines. 
+ * arbitrary long lines.
  */
 #define MAIL_BUFSIZE 1024
 
@@ -128,9 +127,11 @@ smtp_server_t smtp_new(FILE *debug, int protocol)
  */
 
 int smtp_connect(smtp_server_t *srv, const char *host, int port, int timeout,
-	char *errstr)
+	char **server_canonical_name, char **server_address, 
+	char **errstr)
 {
-    return net_open_socket(host, port, timeout, &srv->fd, errstr);
+    return net_open_socket(host, port, timeout, &srv->fd, 
+	    server_canonical_name, server_address, errstr);
 }
 
 
@@ -148,7 +149,7 @@ int smtp_connect(smtp_server_t *srv, const char *host, int port, int timeout,
  * will be returned: SMTP_EIO, SMTP_EPROTO
  */
 
-int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
+int smtp_get_msg(smtp_server_t *srv, list_t **msg, char **errstr)
 {
     list_t *l;
     list_t *lp;
@@ -197,10 +198,21 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 	     * message), but this is the term used in the RFCs.
 	     * An empty reply is a special case of an invalid reply - this
 	     * differentiation may help the user. */
-	    snprintf(errstr, errstr_bufsize, "the server sent an %s reply",
-		    (counter == 0 && len == 0) ? "empty" : "invalid");
+	    if (counter == 0 && len == 0)
+	    {
+		*errstr = xasprintf(_("the server sent an empty reply"));
+	    }
+	    else
+	    {
+		*errstr = xasprintf(_("the server sent an invalid reply"));
+	    }
 	    return SMTP_EPROTO;
 	}
+	if (srv->debug)
+	{
+	    fputs("<-- ", srv->debug);
+	    fwrite(line, sizeof(char), len, srv->debug);
+	}
 	/* kill CRLF */
 	line[--len] = '\0';
 	if (line[len - 1] == '\r')
@@ -210,21 +222,14 @@ int smtp_get_msg(smtp_server_t *srv, list_t **msg, char *errstr)
 	list_insert(lp, xstrdup(line));
 	counter++;
 	lp = lp->next;
-	if (srv->debug)
-	{
-	    fputs("<-- ", srv->debug);
-	    fwrite(line, sizeof(char), len, srv->debug);
-	    fputc('\n', srv->debug);
-	}
     }
     while (line[3] == '-' && counter <= SMTP_MAXLINES);
 
     if (counter > SMTP_MAXLINES)
     {
 	list_xfree(l, free);
-	snprintf(errstr, errstr_bufsize, 
-		"Rejecting server reply that is longer than %d lines. "
-		"Increase SMTP_MAXLINES.", SMTP_MAXLINES);
+	*errstr = xasprintf(_("Rejecting server reply that is longer "
+		    "than %d lines. Increase SMTP_MAXLINES."), SMTP_MAXLINES);
 	return SMTP_EPROTO;
     }
 
@@ -250,12 +255,10 @@ int smtp_msg_status(list_t *msg)
  * smtp_put()
  *
  * This function writes 'len' characters from 's' to the SMTP server 'srv'.
- * If 'newline' is set, it will write TCP CRLF ('\r\n') after the string.
  * Used error codes: SMTP_EIO
  */
 
-int smtp_put(smtp_server_t *srv, const char *s, size_t len, int newline, 
-	char *errstr)
+int smtp_put(smtp_server_t *srv, const char *s, size_t len, char **errstr)
 {
     int e = 0;
 
@@ -263,19 +266,11 @@ int smtp_put(smtp_server_t *srv, const char *s, size_t len, int newline,
     if (tls_is_active(&srv->tls))
     {
 	e = (tls_puts(&srv->tls, s, len, errstr) != TLS_EOK);
-	if (!e && newline)
-	{
-	    e = (tls_puts(&srv->tls, "\r\n", 2, errstr) != TLS_EOK);
-	}
     }
     else
     {
 #endif /* HAVE_SSL */
 	e = (net_puts(srv->fd, s, len, errstr) != NET_EOK);
-	if (!e && newline)
-	{
-	    e = (net_puts(srv->fd, "\r\n", 2, errstr) != NET_EOK);
-	}
 #ifdef HAVE_SSL
     }
 #endif /* HAVE_SSL */
@@ -287,10 +282,6 @@ int smtp_put(smtp_server_t *srv, const char *s, size_t len, int newline,
     {
 	fputs("--> ", srv->debug);
 	fwrite(s, sizeof(char), len, srv->debug);
-	if (newline)
-	{
-	    fputc('\n', srv->debug);
-	}
     }
     
     return SMTP_EOK;
@@ -307,13 +298,15 @@ int smtp_put(smtp_server_t *srv, const char *s, size_t len, int newline,
  * Used error codes: SMTP_EIO, SMTP_EINVAL
  */
 
-/* make gcc print format warnings for this function:
-int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...) 
+/* make gcc print format warnings for this function */
+#ifdef __GNUC__
+int smtp_send_cmd(smtp_server_t *srv, char **errstr, const char *format, ...) 
     __attribute__ ((format (printf, 3, 4)));
-*/
-int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
+#endif
+
+int smtp_send_cmd(smtp_server_t *srv, char **errstr, const char *format, ...)
 {
-    char line[SMTP_MAXCMDLEN + 1];
+    char line[SMTP_MAXCMDLEN + 3];
     int count;
     va_list args;
 
@@ -322,13 +315,15 @@ int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
     va_end(args);
     if (count >= SMTP_MAXCMDLEN + 1)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"Cannot send command because it is "
-		"longer than %d characters. Increase SMTP_MAXCMDLEN.", 
+	*errstr = xasprintf(_("Cannot send command because it is "
+		"longer than %d characters. Increase SMTP_MAXCMDLEN."), 
 		SMTP_MAXCMDLEN);
 	return SMTP_EINVAL;
     }
-    return smtp_put(srv, line, (size_t)count, 1, errstr);
+    line[count++] = '\r';
+    line[count++] = '\n';
+    line[count] = '\0';
+    return smtp_put(srv, line, (size_t)count, errstr);
 }
 
 
@@ -339,7 +334,7 @@ int smtp_send_cmd(smtp_server_t *srv, char *errstr, const char *format, ...)
  */
 
 int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf, 
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *msg;
@@ -352,8 +347,7 @@ int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf,
     if (smtp_msg_status(msg) != 220)
     {
 	*errmsg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"Cannot get initial OK message from server");
+	*errstr = xasprintf(_("cannot get initial OK message from server"));
 	return SMTP_EPROTO;
     }
     if (buf)
@@ -375,7 +369,7 @@ int smtp_get_greeting(smtp_server_t *srv, list_t **errmsg, char **buf,
  */
 
 int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg, 
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *ehlo_response;
@@ -390,27 +384,43 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
     *errmsg = NULL;
     if (srv->protocol == SMTP_PROTO_SMTP)
     {
-	e = smtp_send_cmd(srv, errstr, "EHLO %s", ehlo_domain);
+	if ((e = smtp_send_cmd(srv, errstr, "EHLO %s", ehlo_domain)) 
+		!= SMTP_EOK)
+	{
+	    return e;
+	}
+	if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
+	{
+	    return e;
+	}
+	if (smtp_msg_status(ehlo_response) != 250)
+	{
+	    /* fall back to HELO, for very old SMTP servers */
+	    list_xfree(ehlo_response, free);
+	    if ((e = smtp_send_cmd(srv, errstr, "HELO %s", ehlo_domain)) 
+		    != SMTP_EOK)
+	    {
+		return e;
+	    }
+	    if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
+	    {
+		return e;
+	    }
+	    if (smtp_msg_status(ehlo_response) != 250)
+	    {
+		*errmsg = ehlo_response;
+		*errstr = xasprintf(_("SMTP server does not accept "
+			    "EHLO or HELO commands"));
+		return SMTP_EPROTO;
+	    }
+	    list_xfree(ehlo_response, free);
+	    /* srv->cap.flags is 0 */
+	    return SMTP_EOK;
+	}
     }
-    else
+    else /* protocol is LMTP */
     {
-	e = smtp_send_cmd(srv, errstr, "LHLO %s", ehlo_domain);
-    }
-    if (e != SMTP_EOK)
-    {
-	return e;
-    }
-    if ((e = smtp_get_msg(srv, &ehlo_response, errstr)) != SMTP_EOK)
-    {
-	return e;
-    }
-    if (smtp_msg_status(ehlo_response) != 250)
-    {
-	/* This should only happen with very old SMTP servers that don't support
-	 * EHLO. It should never happen with LMTP. */
-	/* fall back to HELO */
-	list_xfree(ehlo_response, free);
-	if ((e = smtp_send_cmd(srv, errstr, "HELO %s", ehlo_domain)) 
+	if ((e = smtp_send_cmd(srv, errstr, "LHLO %s", ehlo_domain)) 
 		!= SMTP_EOK)
 	{
 	    return e;
@@ -422,13 +432,9 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 	if (smtp_msg_status(ehlo_response) != 250)
 	{
 	    *errmsg = ehlo_response;
-	    snprintf(errstr, errstr_bufsize,
-		    "SMTP server does not accept EHLO or HELO commands");
+	    *errstr = xasprintf(_("command %s failed"), "LHLO");
 	    return SMTP_EPROTO;
 	}
-	list_xfree(ehlo_response, free);
-	/* srv->cap.flags is 0 */
-	return SMTP_EOK;
     }
 
     lp = ehlo_response;
@@ -462,8 +468,8 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 	     * tell us about a maximum message size. Treat that as if the SIZE 
 	     * keyword was not seen. Also treat invalid numbers the same way.
 	     * The value 0 means there is no maximum.
-	     * See RFC 1653.
-	     */
+	     * See RFC 1653. */
+	    errno = 0;
 	    srv->cap.size = strtol(s + 8, &p, 10);
 	    if (!(*(s + 8) == '\0' || *p != '\0' || srv->cap.size < 0
 		    || (srv->cap.size == LONG_MAX && errno == ERANGE)))
@@ -525,7 +531,7 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 
 #ifdef HAVE_SSL
 int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file, 
-	const char *tls_ca_file, const char *tls_trust_file, char *errstr)
+	const char *tls_ca_file, const char *tls_trust_file, char **errstr)
 {
     return tls_init(&srv->tls, tls_key_file, tls_ca_file, tls_trust_file, 
 	    errstr);
@@ -540,7 +546,7 @@ int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file,
  */
 
 #ifdef HAVE_SSL
-int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
+int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -557,8 +563,7 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
     if (smtp_msg_status(msg) != 220)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept the STARTTLS command");
+	*errstr = xasprintf(_("command %s failed"), "STARTTLS");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -575,7 +580,7 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char *errstr)
 
 #ifdef HAVE_SSL
 int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
-	tls_cert_info_t *tci, char *errstr)
+	tls_cert_info_t *tci, char **errstr)
 {
     return tls_start(&srv->tls, srv->fd, hostname, tls_nocertcheck, tci, 
 	    errstr);
@@ -593,7 +598,7 @@ int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
 
 #ifndef USE_GSASL
 int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     char *s;
     char *b64;
@@ -631,15 +636,13 @@ int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
 	*error_msg = msg;
 	if (status == 504)
 	{
-	    snprintf(errstr, errstr_bufsize, 
-		    "AUTH PLAIN not accepted by the server");
+	    *errstr = xasprintf(_("command %s failed"), "AUTH PLAIN");
 	    return SMTP_EPROTO;
 	}
 	else
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "the server does not accept username/password "
-		    "(AUTH PLAIN)");
+	    *errstr = xasprintf(_("authentication failed (method %s)"), 
+		    "PLAIN");
 	    return SMTP_EAUTHFAIL;
 	}
     }
@@ -660,7 +663,7 @@ int smtp_auth_plain(smtp_server_t *srv, const char *user, const char *password,
 
 #ifndef USE_GSASL
 int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -681,8 +684,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH LOGIN not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH LOGIN");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -703,8 +705,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize,
-		"the server does not accept username (AUTH LOGIN)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "LOGIN");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -725,8 +726,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept username/password (AUTH LOGIN)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "LOGIN");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -747,7 +747,7 @@ int smtp_auth_login(smtp_server_t *srv, const char *user, const char *password,
 #ifndef USE_GSASL
 int smtp_auth_cram_md5(smtp_server_t *srv, const char *user, 
 	const char *password,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     unsigned char digest[16];
     char hex[] = "0123456789abcdef";
@@ -774,8 +774,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH CRAM-MD5 not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH CRAM-MD5");
 	return SMTP_EPROTO;
     }
     /* we know the line is at least 4 characters long */
@@ -786,12 +785,12 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (!base64_decode(challenge, challenge_len, b64, &len))
     {
 	list_xfree(msg, free);
-	snprintf(errstr, errstr_bufsize,
-		"AUTH CRAM-MD5: server sent invalid challenge");
+	*errstr = xasprintf(_("authentication method CRAM-MD5: "
+		    "server sent invalid challenge"));
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
-    md5_hmac(password, strlen(password), b64, len, digest);
+    hmac_md5(password, strlen(password), b64, len, digest);
     free(b64);
     
     /* construct username + ' ' + digest_in_hex */
@@ -824,9 +823,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not accept username/password "
-		"(AUTH CRAM-MD5)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "CRAM-MD5");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -848,7 +845,7 @@ int smtp_auth_cram_md5(smtp_server_t *srv, const char *user,
 
 #ifndef USE_GSASL
 int smtp_auth_external(smtp_server_t *srv, const char *user,
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     size_t u_len;
     size_t b64_len;
@@ -868,8 +865,7 @@ int smtp_auth_external(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 334)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"AUTH EXTERNAL not accepted by the server");
+	*errstr = xasprintf(_("command %s failed"), "AUTH EXTERNAL");
 	return SMTP_EPROTO;
     }
     list_xfree(msg, free);
@@ -890,9 +886,7 @@ int smtp_auth_external(smtp_server_t *srv, const char *user,
     if (smtp_msg_status(msg) != 235)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize,
-		"the server does not accept username "
-		"(AUTH EXTERNAL)");
+	*errstr = xasprintf(_("authentication failed (method %s)"), "EXTERNAL");
 	return SMTP_EAUTHFAIL;
     }
     list_xfree(msg, free);
@@ -950,17 +944,10 @@ int smtp_client_supports_authmech(const char *mech)
     
 #else /* not USE_GSASL */
     
-    if (strcmp(mech, "CRAM-MD5") == 0
+    return (strcmp(mech, "CRAM-MD5") == 0
 	    || strcmp(mech, "PLAIN") == 0
 	    || strcmp(mech, "EXTERNAL") == 0
-	    || strcmp(mech, "LOGIN") == 0)
-    {
-	return 1;
-    }
-    else
-    {
-	return 0;
-    }
+	    || strcmp(mech, "LOGIN") == 0);
     
 #endif /* not USE_GSASL */
 }
@@ -976,11 +963,15 @@ int smtp_auth(smtp_server_t *srv,
 	const char *hostname,
 	const char *user, 
 	const char *password,
+#ifdef USE_GSASL
 	const char *ntlmdomain,
+#else
+	const char *ntlmdomain UNUSED,
+#endif
 	const char *auth_mech,
 	char *(*password_callback)(const char *hostname, const char *user),
 	list_t **error_msg,
-	char *errstr)
+	char **errstr)
 {
 #ifdef USE_GSASL
     int e;
@@ -992,29 +983,28 @@ int smtp_auth(smtp_server_t *srv,
     char *outbuf;
     int error_code;
     int auth_plain_special;
-    char *callback_password;
+    char *callback_password = NULL;
 
 
     *error_msg = NULL;
-    if (strcmp(auth_mech, "") != 0 && 
-	    !smtp_server_supports_authmech(srv, auth_mech))
+    if (strcmp(auth_mech, "") != 0 
+	    && !smtp_server_supports_authmech(srv, auth_mech))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not support authentication method %s", 
+	*errstr = xasprintf(
+		_("the server does not support authentication method %s"), 
 		auth_mech);
 	return SMTP_EUNAVAIL;
     }
     if ((error_code = gsasl_init(&ctx)) != GSASL_OK)
     {
-	snprintf(errstr, errstr_bufsize, "GNU SASL: %s", 
-		gsasl_strerror(error_code));
+	*errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	return SMTP_ELIBFAILED;
     }
     if (strcmp(auth_mech, "") != 0 && !gsasl_client_support_p(ctx, auth_mech))
     {
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"GNU SASL: authentication method %s not supported", 
+	*errstr = xasprintf(
+		_("GNU SASL: authentication method %s not supported"), 
 		auth_mech);
 	return SMTP_ELIBFAILED;
     }
@@ -1065,24 +1055,50 @@ int smtp_auth(smtp_server_t *srv,
 	if (!tls_is_active(&srv->tls))
 	{
 #endif /* HAVE_SSL */
-  	    snprintf(errstr, errstr_bufsize,
-   		    "cannot use a secure authentication method");
+	    *errstr = xasprintf(_("cannot use a secure authentication method"));
 #ifdef HAVE_SSL
 	}
 	else
 	{
-   	    snprintf(errstr, errstr_bufsize,
-       		    "cannot find a usable authentication method");
+	    *errstr = xasprintf(
+		    _("cannot find a usable authentication method"));
      	}
 #endif /* not HAVE_SSL */
 	return SMTP_EUNAVAIL;
     }
     
+    /* Check availability of required authentication data */
+    if (strcmp(auth_mech, "EXTERNAL") != 0)
+    {
+	/* GSSAPI, DIGEST-MD5, CRAM-MD5, PLAIN, LOGIN, NTLM all need a user 
+	 * name */
+	if (!user)
+	{
+	    gsasl_done(ctx);
+	    *errstr = xasprintf(_("authentication method %s needs a user name"),
+		    auth_mech);
+	    return SMTP_EUNAVAIL;
+	}
+	/* DIGEST-MD5, CRAM-MD5, PLAIN, LOGIN, NTLM all need a password */
+	if (strcmp(auth_mech, "GSSAPI") != 0 && !password)
+	{
+	    if (!password_callback 
+		    || !(callback_password = password_callback(hostname, user)))
+	    {
+		gsasl_done(ctx);
+		*errstr = xasprintf(
+			_("authentication method %s needs a password"),
+			auth_mech);
+		return SMTP_EUNAVAIL;
+	    }
+	    password = callback_password;
+	}
+    }
+
     if ((error_code = gsasl_client_start(ctx, auth_mech, &sctx)) != GSASL_OK)
     {
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"GNU SASL: %s", gsasl_strerror(error_code));
+	*errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	return SMTP_ELIBFAILED;
     }
 
@@ -1102,22 +1118,7 @@ int smtp_auth(smtp_server_t *srv,
     {
 	gsasl_property_set(sctx, GSASL_PASSWORD, password);
     }
-    /* If the callback fails, or there is none, leave the error message about a
-     * missing password to GSASL. */
-    else if (password_callback)
-    {
-	if (strcmp(auth_mech, "GSSAPI") != 0
-		&& strcmp(auth_mech, "EXTERNAL") != 0)
-	{
-	    /* All others (PLAIN, LOGIN, CRAM-MD5, DIGEST-MD5, NTLM) need a
-	     * password. */
-	    if ((callback_password = password_callback(hostname, user)))
-	    {
-		gsasl_property_set(sctx, GSASL_PASSWORD, callback_password);
-		free(callback_password);
-	    }	    
-	}
-    }
+    free(callback_password);
     /* For DIGEST-MD5 and GSSAPI */
     gsasl_property_set(sctx, GSASL_SERVICE, "smtp");
     if (hostname)
@@ -1140,8 +1141,7 @@ int smtp_auth(smtp_server_t *srv,
 	{
 	    gsasl_finish(sctx);
 	    gsasl_done(ctx);
-	    snprintf(errstr, errstr_bufsize, 
-		    "GNU SASL: %s", gsasl_strerror(error_code));
+	    *errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
 	    return SMTP_ELIBFAILED;
 	}
 	if (!input)
@@ -1187,8 +1187,7 @@ int smtp_auth(smtp_server_t *srv,
 		gsasl_finish(sctx);
 		gsasl_done(ctx);
 		free(outbuf);
-		snprintf(errstr, errstr_bufsize, 
-			"authentication failed (GNU SASL, method %s)", 
+		*errstr = xasprintf(_("authentication failed (method %s)"), 
 			auth_mech);
 		return SMTP_EAUTHFAIL;
 	    }	    
@@ -1227,8 +1226,7 @@ int smtp_auth(smtp_server_t *srv,
 		gsasl_finish(sctx);
 		gsasl_done(ctx);
 		free(outbuf);
-		snprintf(errstr, errstr_bufsize, 
-			"authentication failed (GNU SASL, method %s)", 
+		*errstr = xasprintf(_("authentication failed (method %s)"), 
 			auth_mech);
 		return SMTP_EAUTHFAIL;
 	    }	    
@@ -1244,8 +1242,7 @@ int smtp_auth(smtp_server_t *srv,
     {
 	gsasl_finish(sctx);
 	gsasl_done(ctx);
-	snprintf(errstr, errstr_bufsize, 
-		"authentication failed: %s (GNU SASL, method %s)", 
+	*errstr = xasprintf(_("authentication failed: %s (method %s)"), 
 		gsasl_strerror(error_code), auth_mech);
 	return SMTP_EAUTHFAIL;
     }
@@ -1266,8 +1263,7 @@ int smtp_auth(smtp_server_t *srv,
 	if (smtp_msg_status(msg) != 235)
 	{
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize, 
-	    	    "authentication failed (GNU SASL, method %s)", 
+	    *errstr = xasprintf(_("authentication failed (method %s)"),
 		    auth_mech);
 	    return SMTP_EAUTHFAIL;
 	}
@@ -1282,11 +1278,11 @@ int smtp_auth(smtp_server_t *srv,
     
     
     *error_msg = NULL;
-    if (strcmp(auth_mech, "") != 0 && 
-	    !smtp_server_supports_authmech(srv, auth_mech))
+    if (strcmp(auth_mech, "") != 0 
+	    && !smtp_server_supports_authmech(srv, auth_mech))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"the server does not support authentication method %s", 
+	*errstr = xasprintf( 
+		_("the server does not support authentication method %s"), 
 		auth_mech);
 	return SMTP_EUNAVAIL;
     }
@@ -1317,14 +1313,13 @@ int smtp_auth(smtp_server_t *srv,
        	if (!tls_is_active(&srv->tls))
  	{
 #endif /* HAVE_SSL */
-  	    snprintf(errstr, errstr_bufsize,
-   		    "cannot use a secure authentication method");
+	    *errstr = xasprintf(_("cannot use a secure authentication method"));
 #ifdef HAVE_SSL
     	}
     	else
  	{
-      	    snprintf(errstr, errstr_bufsize,
-      		    "cannot find a usable authentication method");
+	    *errstr = xasprintf(
+		    _("cannot find a usable authentication method"));
 	}
 #endif /* not HAVE_SSL */
 	return SMTP_EUNAVAIL;
@@ -1335,8 +1330,8 @@ int smtp_auth(smtp_server_t *srv,
 	/* CRAMD-MD5, PLAIN, LOGIN all need a user name and a password */
 	if (!user)
 	{
-	    snprintf(errstr, errstr_bufsize,
-	    	    "authentication method %s needs a user name", auth_mech);
+	    *errstr = xasprintf(_("authentication method %s needs a user name"),
+		    auth_mech);
 	    return SMTP_EUNAVAIL;
 	}
 	if (!password)
@@ -1344,8 +1339,8 @@ int smtp_auth(smtp_server_t *srv,
 	    if (!password_callback 
 		    || !(callback_password = password_callback(hostname, user)))
 	    {
-		snprintf(errstr, errstr_bufsize,
-		    	"authentication method %s needs a password", 
+		*errstr = xasprintf(
+			_("authentication method %s needs a password"),
 			auth_mech);
 		return SMTP_EUNAVAIL;
 	    }
@@ -1371,8 +1366,8 @@ int smtp_auth(smtp_server_t *srv,
     }
     else
     {
-	snprintf(errstr, errstr_bufsize, 
-		"authentication method %s not supported", auth_mech);
+	*errstr = xasprintf(_("authentication method %s not supported"),
+		auth_mech);
 	e = SMTP_ELIBFAILED;
     }
     free(callback_password);
@@ -1394,7 +1389,7 @@ int smtp_send_envelope(smtp_server_t *srv,
 	const char *dsn_notify,
 	const char *dsn_return,
 	list_t **error_msg,
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *msg;    
@@ -1481,9 +1476,8 @@ int smtp_send_envelope(smtp_server_t *srv,
 		if (smtp_msg_status(msg) != 250)
 		{
 		    *error_msg = msg;
-		    snprintf(errstr, errstr_bufsize, 
-			    "envelope from address %s not accepted "
-			    "by the server", envelope_from);
+		    *errstr = xasprintf(_("envelope from address %s not "
+				"accepted by the server"), envelope_from);
 		    return SMTP_EINVAL;
 		}
 		list_xfree(msg, free);
@@ -1499,8 +1493,8 @@ int smtp_send_envelope(smtp_server_t *srv,
 		if (smtp_msg_status(msg) != 250)
 		{
 		    *error_msg = msg;
-		    snprintf(errstr, errstr_bufsize, 
-			    "recipient address %s not accepted by the server",
+		    *errstr = xasprintf(_("recipient address %s not "
+				"accepted by the server"),
 			    (char *)(rcpt_recv->data));
 		    return SMTP_EINVAL;
 		}
@@ -1515,8 +1509,8 @@ int smtp_send_envelope(smtp_server_t *srv,
 		if (smtp_msg_status(msg) != 354)
 		{
 		    *error_msg = msg;
-		    snprintf(errstr, errstr_bufsize, 
-			    "the server does not accept mail data");
+		    *errstr = xasprintf(
+			    _("the server does not accept mail data"));
 		    return SMTP_EUNAVAIL;
 		}
 		list_xfree(msg, free);
@@ -1537,23 +1531,28 @@ int smtp_send_envelope(smtp_server_t *srv,
  */
 
 int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc, 
-	long *mailsize, char *errstr)
+	long *mailsize, char **errstr)
 {
-    char buffer[MAIL_BUFSIZE];
+    char bigbuffer[MAIL_BUFSIZE + 3];	/* buffer + leading dot + ending CRLF */
+    char *buffer;
     size_t len;
+    char *send_buf;
+    size_t send_len;
     int in_header;
     int in_bcc;
     int line_starts;
     int line_continues;
     int e;
     
+    bigbuffer[0] = '.';
+    buffer = bigbuffer + 1;
     in_header = 1;
     in_bcc = 0;
     line_continues = 0;
     e = SMTP_EOK;
     for (;;)
     {
-	if (stream_gets(mailf, buffer, sizeof(buffer), &len, errstr) 
+	if (stream_gets(mailf, buffer, MAIL_BUFSIZE, &len, errstr) 
 		!= STREAM_EOK)
 	{
 	    return SMTP_EIO;
@@ -1636,18 +1635,26 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
 		}
 	    }
 	}
+	send_buf = buffer;
+	send_len = len;
 	if (line_starts && buffer[0] == '.')
 	{
-	    e = smtp_put(srv, ".", 1, 0, errstr);
-	    /* do not count the quoting dot in 'mailsize' */
+	    /* Quote the leading dot with another dot */
+	    send_buf = bigbuffer;
+	    send_len = len + 1;
 	}
-	if (e == SMTP_EOK)
+	if (!line_continues)
 	{
-    	    e = smtp_put(srv, buffer, len, !line_continues, errstr);
-	    /* count a newline as one character */
-	    *mailsize += (long)len + (line_continues ? 0 : 1);
+	    /* Append CRLF */
+	    buffer[len] = '\r';
+	    buffer[len + 1] = '\n';
+	    buffer[len + 2] = '\0';
+	    send_len += 2;
 	}
-    	if (e != SMTP_EOK)
+	/* Update mailsize. Do not count a quote dot. Count CRLF as one
+	 * character. */
+	*mailsize += (long)len + (line_continues ? 0 : 1);
+	if ((e = smtp_put(srv, send_buf, send_len, errstr)) != SMTP_EOK)
 	{
 	    return e;
 	}
@@ -1663,7 +1670,7 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
  * see smtp.h
  */
 
-int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char *errstr)
+int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -1680,8 +1687,7 @@ int smtp_end_mail(smtp_server_t *srv, list_t **error_msg, char *errstr)
     if (smtp_msg_status(msg) != 250)
     {
 	*error_msg = msg;
-	snprintf(errstr, errstr_bufsize, 
-		"the server did not accept the mail");
+	*errstr = xasprintf(_("the server did not accept the mail"));
 	return SMTP_EUNAVAIL;
     }
     list_xfree(msg, free);
@@ -1706,7 +1712,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
 	list_t *recipients, 
 	list_t **errstrs,
 	list_t **error_msgs,
-	char *errstr)
+	char **errstr)
 {
     int e;
     list_t *msg;
@@ -1744,9 +1750,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
 	if (smtp_msg_status(msg) != 250)
 	{
 	    all_recipients_accepted = 0;
-	    tmp = xmalloc(errstr_bufsize * sizeof(char));
-	    snprintf(tmp, errstr_bufsize, 
-		    "the server refuses to send the mail to %s",
+	    tmp = xasprintf(_("the server refuses to send the mail to %s"),
 		    (char *)(lp_recipients->data));
 	    list_insert(lp_errstrs, tmp);
 	    list_insert(lp_error_msgs, msg);
@@ -1784,7 +1788,7 @@ int smtp_end_mail_lmtp(smtp_server_t *srv,
  */
 
 int smtp_etrn(smtp_server_t *srv, const char *etrn_argument, 
-	list_t **error_msg, char *errstr)
+	list_t **error_msg, char **errstr)
 {
     int e;
     list_t *msg;
@@ -1809,23 +1813,22 @@ int smtp_etrn(smtp_server_t *srv, const char *etrn_argument,
 	case 458: /* Unable to queue messages for node <x> */
 	case 459: /* Node <x> not allowed: <reason> */
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "the server is unable to fulfill the request");
+	    *errstr = xasprintf(
+		    _("the server is unable to fulfill the request"));
 	    return SMTP_EUNAVAIL;
 	    break;
 	    
 	case 500: /* Syntax Error */
 	case 501: /* 501 Syntax Error in Parameters */
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "invalid argument for Remote Message Queue Starting");
+	    *errstr = xasprintf(
+		    _("invalid argument for Remote Message Queue Starting"));
 	    return SMTP_EINVAL;
 	    break;
 	    
 	default:
 	    *error_msg = msg;
-	    snprintf(errstr, errstr_bufsize,
-		    "the server sent an invalid reply to the ETRN command");
+	    *errstr = xasprintf(_("command %s failed"), "ETRN");
 	    return SMTP_EPROTO;
 	    break;
     }
@@ -1841,7 +1844,7 @@ int smtp_etrn(smtp_server_t *srv, const char *etrn_argument,
  * see smtp.h
  */
 
-int smtp_quit(smtp_server_t *srv, char *errstr)
+int smtp_quit(smtp_server_t *srv, char **errstr)
 {
     int e;
     list_t *msg;
