@@ -3,7 +3,7 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -17,11 +17,8 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with this program; if not, write to the Free Software
- *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- *
- *   msmtp is released under the GPL with the additional exemption that
- *   compiling, linking, and/or using OpenSSL is allowed.
+ *   along with this program; if not, write to the Free Software Foundation,
+ *   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -51,7 +48,13 @@ extern int errno;
 #include <openssl/evp.h>
 #endif /* HAVE_OPENSSL */
 
+#ifdef USE_LIBIDN
+#include <idna.h>
+#endif
+
+#include "gettext.h"
 #include "xalloc.h"
+#include "xvasprintf.h"
 
 #include "tls.h"
 
@@ -77,7 +80,7 @@ void tls_clear(tls_t *tls)
  */
 
 #ifdef HAVE_OPENSSL
-int seed_prng(char *errstr)
+int seed_prng(char **errstr)
 {
     char randfile[512];
     time_t t;
@@ -92,22 +95,21 @@ int seed_prng(char *errstr)
     {
 	if (!RAND_file_name(randfile, 512))
 	{
-	    snprintf(errstr, errstr_bufsize, 
-	    	    "no environment variables RANDFILE or HOME, "
-	    	    "or filename of rand file too long");
+	    *errstr = xasprintf(_("no environment variables RANDFILE or HOME, "
+			"or filename of rand file too long"));
 	    return TLS_ESEED;
 	}
 	if (RAND_load_file(randfile, -1) < 1)
 	{
-	    snprintf(errstr, errstr_bufsize, "error reading %s", randfile);
+	    *errstr = xasprintf(_("%s: input error"), randfile);
 	    return TLS_ESEED;
 	}
 	/* Seed in time. I can't think of other "random" things on DOS
 	 * systems. */
 	if ((t = time(NULL)) < 0)
 	{
-	    snprintf(errstr, errstr_bufsize, "cannot get system time: %s",
-	    	    strerror(errno));
+	    *errstr = xasprintf(_("cannot get system time: %s"), 
+		    strerror(errno));
 	    return TLS_ESEED;
 	}
 	RAND_seed((unsigned char *)&t, sizeof(time_t));
@@ -126,8 +128,8 @@ int seed_prng(char *errstr)
 	/* Are we happy now? */
 	if (!RAND_status())
 	{
-	    snprintf(errstr, errstr_bufsize, "random file + time + pseudo "
-		    "randomness is not enough, giving up");
+	    *errstr = xasprintf(_("random file + time + pseudo randomness is "
+			"not enough, giving up"));
 	    return TLS_ESEED;
 	}
 	/* Save a rand file for later usage. We ignore errors here as we can't
@@ -145,15 +147,14 @@ int seed_prng(char *errstr)
  * see tls.h
  */
 
-int tls_lib_init(char *errstr)
+int tls_lib_init(char **errstr)
 {
 #ifdef HAVE_GNUTLS
     int error_code;
     
     if ((error_code = gnutls_global_init()) != 0)
     {
-	snprintf(errstr, errstr_bufsize, "%s", 
-		gnutls_strerror(error_code));
+	*errstr = xasprintf("%s", gnutls_strerror(error_code));
 	return TLS_ELIBFAILED;
     }
 
@@ -350,7 +351,7 @@ error_exit:
  * see tls.h
  */
 
-int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
+int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char **errstr)
 {
 #ifdef HAVE_GNUTLS
     const gnutls_datum_t *cert_list;
@@ -366,26 +367,26 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
     int i;	
     int e;
     char *p;
-    const char *errmsg = "cannot get TLS certificate info";
-
+    const char *errmsg;
+    
+    errmsg = _("cannot get TLS certificate info");
     if (!(cert_list = 
 		gnutls_certificate_get_peers(tls->session, &cert_list_size))
 	    || cert_list_size == 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: no certificate was found", errmsg);
+	*errstr = xasprintf(_("%s: no certificate was found"), errmsg);
 	return TLS_ECERT;
     }
     if (gnutls_x509_crt_init(&cert) != 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot initialize certificate structure", errmsg);
+	*errstr = xasprintf(_("%s: cannot initialize certificate structure"),
+		errmsg);
 	return TLS_ECERT;
     }
     if (gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER) != 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: error parsing certificate", errmsg);
+	*errstr = xasprintf(_("%s: error parsing certificate"), errmsg);
+	gnutls_x509_crt_deinit(cert);
 	return TLS_ECERT;
     }
 
@@ -394,28 +395,28 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
     if (gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_SHA, 
 		tci->sha1_fingerprint, &size) != 0)
     {
-	snprintf(errstr, errstr_bufsize,
-		"%s: error getting SHA1 fingerprint", errmsg);
+	*errstr = xasprintf(_("%s: error getting SHA1 fingerprint"), errmsg);
+	gnutls_x509_crt_deinit(cert);
 	return TLS_ECERT;
     }
     size = 16;
     if (gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_MD5, 
 		tci->md5_fingerprint, &size) != 0)
     {
-	snprintf(errstr, errstr_bufsize,
-		"%s: error getting MD5 fingerprint", errmsg);
+	*errstr = xasprintf(_("%s: error getting MD5 fingerprint"), errmsg);
+	gnutls_x509_crt_deinit(cert);
 	return TLS_ECERT;
     }
     if ((tci->activation_time = gnutls_x509_crt_get_activation_time(cert)) < 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get activation time", errmsg);
+	*errstr = xasprintf(_("%s: cannot get activation time"), errmsg);
+	gnutls_x509_crt_deinit(cert);
 	return TLS_ECERT;
     }
     if ((tci->expiration_time = gnutls_x509_crt_get_expiration_time(cert)) < 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get expiration time", errmsg);
+	*errstr = xasprintf(_("%s: cannot get expiration time"), errmsg);
+	gnutls_x509_crt_deinit(cert);
 	return TLS_ECERT;
     }
 
@@ -461,6 +462,7 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
 	}
     }
     
+    gnutls_x509_crt_deinit(cert);
     return TLS_EOK;
 #endif /* HAVE_GNUTLS */
     
@@ -479,26 +481,24 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
     unsigned int usize;
     char *p;
     int i;
-    const char *errmsg = "cannot get TLS certificate info";
-
+    const char *errmsg;
+    
+    errmsg = _("cannot get TLS certificate info");
     if (!(x509cert = SSL_get_peer_certificate(tls->ssl)))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: no certificate was found", errmsg);
+	*errstr = xasprintf(_("%s: no certificate was found"), errmsg);
 	return TLS_ECERT;
     }
     if (!(x509_subject = X509_get_subject_name(x509cert)))
     {
+	*errstr = xasprintf(_("%s: cannot get certificate subject"), errmsg);
 	X509_free(x509cert);
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get certificate subject", errmsg);
 	return TLS_ECERT;
     }
     if (!(x509_issuer = X509_get_issuer_name(x509cert)))
     {
+	*errstr = xasprintf(_("%s: cannot get certificate issuer"), errmsg);
 	X509_free(x509cert);
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get certificate issuer", errmsg);
 	return TLS_ECERT;
     }
 
@@ -506,15 +506,13 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
     usize = 20;
     if (!X509_digest(x509cert, EVP_sha1(), tci->sha1_fingerprint, &usize))
     {
-	snprintf(errstr, errstr_bufsize,
-		"%s: error getting SHA1 fingerprint", errmsg);
+	*errstr = xasprintf(_("%s: error getting SHA1 fingerprint"), errmsg);
 	return TLS_ECERT;
     }
     usize = 16;
     if (!X509_digest(x509cert, EVP_md5(), tci->md5_fingerprint, &usize))
     {
-	snprintf(errstr, errstr_bufsize,
-		"%s: error getting MD5 fingerprint", errmsg);
+	*errstr = xasprintf(_("%s: error getting MD5 fingerprint"), errmsg);
 	return TLS_ECERT;
     }
     asn1time = X509_get_notBefore(x509cert);
@@ -522,10 +520,9 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
 		(asn1time->type != V_ASN1_GENERALIZEDTIME), 
 		&(tci->activation_time)) != 0)
     {
+	*errstr = xasprintf(_("%s: cannot get activation time"), errmsg);
 	X509_free(x509cert);
 	tls_cert_info_free(tci);
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get activation time", errmsg);
 	return TLS_ECERT;
     }
     asn1time = X509_get_notAfter(x509cert);
@@ -533,10 +530,9 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char *errstr)
 		(asn1time->type != V_ASN1_GENERALIZEDTIME), 
 		&(tci->expiration_time)) != 0)
     {
+	*errstr = xasprintf(_("%s: cannot get expiration time"), errmsg);
 	X509_free(x509cert);
 	tls_cert_info_free(tci);
-	snprintf(errstr, errstr_bufsize, 
-		"%s: cannot get expiration time", errmsg);
 	return TLS_ECERT;
     }
 
@@ -648,7 +644,7 @@ int hostname_match(const char *hostname, const char *certname)
  * Used error codes: TLS_ECERT
  */
 
-int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
+int tls_check_cert(tls_t *tls, const char *hostname, int verify, char **errstr)
 {
 #ifdef HAVE_GNUTLS
     int error_code;
@@ -659,65 +655,69 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
     unsigned int i;
     gnutls_x509_crt_t cert;
     time_t t1, t2;
+#ifdef USE_LIBIDN
+    char *hostname_ascii;
+#endif
     
     if (verify)
     {
-	error_msg = "TLS certificate verification failed";
+	error_msg = _("TLS certificate verification failed");
     }
     else
     {
-	error_msg = "TLS certificate check failed";
+	error_msg = _("TLS certificate check failed");
     }
     
     /* If 'verify' is true, this function uses the trusted CAs in the 
      * credentials structure. So you must have installed one or more CA 
      * certificates. */
+    gnutls_certificate_set_verify_flags(tls->cred, 
+	    GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
     if ((error_code = gnutls_certificate_verify_peers2(tls->session, 
 		    &status)) != 0)
     {
-	snprintf(errstr, errstr_bufsize,
-		"%s: %s", error_msg, gnutls_strerror(error_code));
+	*errstr = xasprintf("%s: %s", error_msg, gnutls_strerror(error_code));
 	return TLS_ECERT;
     }
     if (verify)
     {
 	if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
 	{
-	    snprintf(errstr, errstr_bufsize, 
-	    	    "%s: the certificate hasn't got a known issuer", error_msg);
+	    *errstr = xasprintf(
+		    _("%s: the certificate hasn't got a known issuer"),
+		    error_msg);
 	    return TLS_ECERT;
 	}
 	if (status & GNUTLS_CERT_INVALID)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: the certificate is not trusted", error_msg);
+	    *errstr = xasprintf(_("%s: the certificate is not trusted"),
+		    error_msg);
 	    return TLS_ECERT;
 	}
     }
     if (status & GNUTLS_CERT_REVOKED)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: the certificate has been revoked", error_msg);
+	*errstr = xasprintf(_("%s: the certificate has been revoked"),
+		error_msg);
 	return TLS_ECERT;
     }
     if (gnutls_certificate_type_get(tls->session) != GNUTLS_CRT_X509)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: the certificate type is not X509", error_msg);
+	*errstr = xasprintf(_("%s: the certificate type is not X509"),
+		error_msg);
 	return TLS_ECERT;
     }
     if (!(cert_list = gnutls_certificate_get_peers(
 		    tls->session, &cert_list_size)))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: no certificate was found", error_msg);
+	*errstr = xasprintf(_("%s: no certificate was found"), error_msg);
 	return TLS_ECERT;
     }
     /* Needed to check times: */
     if ((t1 = time(NULL)) < 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: %s", error_msg, strerror(errno));
+	*errstr = xasprintf("%s: cannot get system time: %s", error_msg, 
+		strerror(errno));
 	return TLS_ECERT;
     }
     /* Check the certificate chain. All certificates in the chain must have
@@ -727,52 +727,67 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
     {
 	if (gnutls_x509_crt_init(&cert) < 0)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: cannot initialize certificate structure", error_msg);
+	    *errstr = xasprintf(
+		    _("%s: cannot initialize certificate structure"), 
+		    error_msg);
 	    return TLS_ECERT;
 	}
 	if (gnutls_x509_crt_import(cert, &cert_list[i], GNUTLS_X509_FMT_DER) 
 		< 0)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: error parsing certificate %u of %u", 
+	    *errstr = xasprintf(_("%s: error parsing certificate %u of %u"), 
 		    error_msg, i + 1, cert_list_size);
 	    return TLS_ECERT;
 	}
 	/* Check hostname */
-	if (i == 0 && !gnutls_x509_crt_check_hostname(cert, hostname)) 
+	if (i == 0)
 	{
-    	    snprintf(errstr, errstr_bufsize, 
-		    "%s: the certificate owner does not match hostname %s", 
-		    error_msg, hostname);
-    	    return TLS_ECERT; 
+#ifdef USE_LIBIDN
+	    if (idna_to_ascii_lz(hostname, &hostname_ascii, 0) == IDNA_SUCCESS)
+	    {
+    		if (!gnutls_x509_crt_check_hostname(cert, hostname_ascii)) 
+    		{
+    		    *errstr = xasprintf(_("%s: the certificate owner does not "
+				"match hostname %s"), error_msg, hostname);
+		    free(hostname_ascii);
+	    	    return TLS_ECERT; 
+		}
+    		free(hostname_ascii);
+	    }
+	    else
+#endif
+	    if (!gnutls_x509_crt_check_hostname(cert, hostname)) 
+	    {
+	    	*errstr = xasprintf(_("%s: the certificate owner does not "
+    			    "match hostname %s"), error_msg, hostname);
+		return TLS_ECERT; 
+	    }
 	}
 	/* Check certificate times */
 	if ((t2 = gnutls_x509_crt_get_activation_time(cert)) < 0)
 	{
-	    snprintf(errstr, errstr_bufsize, 
-		    "%s: cannot get activation time for certificate "
-		    "%u of %u", error_msg, i + 1, cert_list_size);
+	    *errstr = xasprintf(_("%s: cannot get activation time for "
+			"certificate %u of %u"), 
+		    error_msg, i + 1, cert_list_size);
 	    return TLS_ECERT; 
 	}
 	if (t2 > t1)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: certificate %u of %u is not yet activated", 
+	    *errstr = xasprintf(
+		    _("%s: certificate %u of %u is not yet activated"),
 		    error_msg, i + 1, cert_list_size);
 	    return TLS_ECERT;
 	}
 	if ((t2 = gnutls_x509_crt_get_expiration_time(cert)) < 0)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: cannot get expiration time for certificate "
-		    "%u of %u", error_msg, i + 1, cert_list_size);
+	    *errstr = xasprintf(_("%s: cannot get expiration time for "
+			"certificate %u of %u"), 
+		    error_msg, i + 1, cert_list_size);
 	    return TLS_ECERT;
 	}
 	if (t2 < t1)
 	{
-	    snprintf(errstr, errstr_bufsize,
-		    "%s: certificate %u of %u has expired", 
+	    *errstr = xasprintf(_("%s: certificate %u of %u has expired"), 
 		    error_msg, i + 1, cert_list_size);
 	    return TLS_ECERT;
 	}
@@ -787,6 +802,8 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
     long status;
     const char *error_msg;
     int i;
+    /* hostname in ASCII format: */
+    char *hostname_ascii;
     /* needed to get the common name: */
     X509_NAME *x509_subject;
     char *buf;
@@ -801,18 +818,17 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
 
     if (verify)
     {
-	error_msg = "TLS certificate verification failed";
+	error_msg = _("TLS certificate verification failed");
     }
     else
     {
-	error_msg = "TLS certificate check failed";
+	error_msg = _("TLS certificate check failed");
     }
 
     /* Get certificate */
     if (!(x509cert = SSL_get_peer_certificate(tls->ssl)))
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: no certificate was sent", error_msg);
+	*errstr = xasprintf(_("%s: no certificate was sent"), error_msg);
 	return TLS_ECERT;
     }
     
@@ -823,9 +839,9 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
 		    && status != X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
 		    && status != X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN))
 	{
+	    *errstr = xasprintf("%s: %s", error_msg, 
+		    X509_verify_cert_error_string(status));
 	    X509_free(x509cert);
-    	    snprintf(errstr, errstr_bufsize,
-		    "%s: %s", error_msg, X509_verify_cert_error_string(status));
 	    return TLS_ECERT;
 	}
     }
@@ -833,6 +849,15 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
     /* Check if 'hostname' matches the one of the subjectAltName extensions of
      * type DNS or the Common Name (CN). */
     
+#ifdef USE_LIBIDN
+    if (idna_to_ascii_lz(hostname, &hostname_ascii, 0) != IDNA_SUCCESS)
+    {
+	hostname_ascii = xstrdup(hostname);
+    }
+#else
+    hostname_ascii = xstrdup(hostname);
+#endif
+    
     /* Try the DNS subjectAltNames. */
     match_found = 0;
     if ((subj_alt_names = 
@@ -844,7 +869,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
 	    subj_alt_name = sk_GENERAL_NAME_value(subj_alt_names, i);
 	    if (subj_alt_name->type == GEN_DNS) 
 	    {
-		if ((match_found = hostname_match(hostname, 
+		if ((match_found = hostname_match(hostname_ascii, 
 				(char *)(subj_alt_name->d.ia5->data))))
 		{
 		    break;
@@ -857,9 +882,9 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
 	/* Try the common name */    
 	if (!(x509_subject = X509_get_subject_name(x509cert)))
 	{
+	    *errstr = xasprintf(_("%s: cannot get certificate subject"),
+		    error_msg);
 	    X509_free(x509cert);
-	    snprintf(errstr, errstr_bufsize,
-	    	    "%s: cannot get certificate subject", error_msg);
 	    return TLS_ECERT;
 	}
 	bufsize = X509_NAME_get_text_by_NID(x509_subject, NID_commonName, 
@@ -869,21 +894,22 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
 	if (X509_NAME_get_text_by_NID(x509_subject, NID_commonName, 
 		    buf, bufsize) == -1)
 	{
+	    *errstr = xasprintf(_("%s: cannot get certificate common name"),
+		    error_msg);
 	    X509_free(x509cert);
 	    free(buf);
-	    snprintf(errstr, errstr_bufsize, 
-		    "%s: cannot get certificate common name", error_msg);
 	    return TLS_ECERT;
 	}
-	match_found = hostname_match(hostname, buf);
+	match_found = hostname_match(hostname_ascii, buf);
 	free(buf);
     }
     X509_free(x509cert);
+    free(hostname_ascii);
 
     if (!match_found)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"%s: the certificate owner does not match hostname %s", 
+	*errstr = xasprintf(
+		_("%s: the certificate owner does not match hostname %s"), 
 		error_msg, hostname);
 	return TLS_ECERT;
     }
@@ -900,32 +926,30 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char *errstr)
  */
 
 int tls_init(tls_t *tls, const char *key_file, const char *cert_file, 
-	const char *trust_file, char *errstr)
+	const char *trust_file, char **errstr)
 {
 #ifdef HAVE_GNUTLS
     int error_code;
     
     if ((error_code = gnutls_init(&tls->session, GNUTLS_CLIENT)) != 0)
     {
-	snprintf(errstr, errstr_bufsize, 
-		"cannot initialize TLS Session: %s", 
+	*errstr = xasprintf(_("cannot initialize TLS Session: %s"), 
 		gnutls_strerror(error_code));
 	return TLS_ELIBFAILED;
     }
     if ((error_code = gnutls_set_default_priority(tls->session)) != 0)
     {
-	gnutls_deinit(tls->session);
-	snprintf(errstr, errstr_bufsize, 
-		"cannot set priorities on TLS Session: %s",
+	*errstr = xasprintf(_("cannot set priorities on TLS Session: %s"),
 		gnutls_strerror(error_code));
+	gnutls_deinit(tls->session);
 	return TLS_ELIBFAILED;
     }
     if ((error_code = gnutls_certificate_allocate_credentials(&tls->cred)) < 0)
     {
-	gnutls_deinit(tls->session);
-	snprintf(errstr, errstr_bufsize, 
-		"cannot allocate certificate for TLS Session: %s",
+	*errstr = xasprintf(
+		_("cannot allocate certificate for TLS Session: %s"),
 		gnutls_strerror(error_code));
+	gnutls_deinit(tls->session);
 	return TLS_ELIBFAILED;
     }
     if (key_file && cert_file)
@@ -933,12 +957,11 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 	if ((error_code = gnutls_certificate_set_x509_key_file(tls->cred, 
 		       	cert_file, key_file, GNUTLS_X509_FMT_PEM)) < 0)
 	{
+	    *errstr = xasprintf(_("cannot set X509 key file %s and/or "
+			"X509 cert file %s for TLS Session: %s"),
+	    	    key_file, cert_file, gnutls_strerror(error_code));
 	    gnutls_deinit(tls->session);
 	    gnutls_certificate_free_credentials(tls->cred);
-	    snprintf(errstr, errstr_bufsize, 
-		    "cannot set X509 key file %s and/or "
-		    "X509 cert file %s for TLS Session: %s",
-	    	    key_file, cert_file, gnutls_strerror(error_code));
 	    return TLS_EFILE;
 	}
     }
@@ -947,11 +970,11 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 	if ((error_code = gnutls_certificate_set_x509_trust_file(
 			tls->cred, trust_file, GNUTLS_X509_FMT_PEM)) <= 0)
 	{
+	    *errstr = xasprintf(
+		    _("cannot set X509 trust file %s for TLS Session: %s"),
+	    	    trust_file, gnutls_strerror(error_code));
 	    gnutls_deinit(tls->session);
 	    gnutls_certificate_free_credentials(tls->cred);
-	    snprintf(errstr, errstr_bufsize, 
-		    "cannot set X509 trust file %s for TLS Session: %s",
-	    	    trust_file, gnutls_strerror(error_code));
 	    return TLS_EFILE;
 	}
 	tls->have_trust_file = 1;
@@ -959,11 +982,10 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
     if ((error_code = gnutls_credentials_set(tls->session, 
 		    GNUTLS_CRD_CERTIFICATE, tls->cred)) < 0)
     {
+	*errstr = xasprintf(_("cannot set credentials for TLS Session: %s"),
+		gnutls_strerror(error_code));
 	gnutls_deinit(tls->session);
 	gnutls_certificate_free_credentials(tls->cred);
-	snprintf(errstr, errstr_bufsize, 
-		"cannot set credentials for TLS Session: %s",
-		gnutls_strerror(error_code));
 	return TLS_ELIBFAILED;
     }
     return TLS_EOK;
@@ -976,13 +998,13 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
     
     if (!(ssl_method = SSLv23_client_method()))
     {
-	snprintf(errstr, errstr_bufsize, "cannot set TLS method");
+	*errstr = xasprintf(_("cannot set TLS method"));
 	return TLS_ELIBFAILED;
     }
     tls->ssl_ctx = SSL_CTX_new(ssl_method);
     if (!tls->ssl_ctx)
     {
-	snprintf(errstr, errstr_bufsize, "cannot create TLS context: %s",
+	*errstr = xasprintf(_("cannot create TLS context: %s"),
 		ERR_error_string(ERR_get_error(), NULL));
 	return TLS_ELIBFAILED;
     }
@@ -991,20 +1013,18 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 	if (SSL_CTX_use_PrivateKey_file(
 		    tls->ssl_ctx, key_file, SSL_FILETYPE_PEM) != 1)
 	{
+	    *errstr = xasprintf(_("cannot load key file %s: %s"), 
+		    key_file, ERR_error_string(ERR_get_error(), NULL));
 	    SSL_CTX_free(tls->ssl_ctx);
 	    tls->ssl_ctx = NULL;
-	    snprintf(errstr, errstr_bufsize,
-		    "cannot load key file %s: %s", 
-		    key_file, ERR_error_string(ERR_get_error(), NULL));
 	    return TLS_EFILE;
 	}
 	if (SSL_CTX_use_certificate_chain_file(tls->ssl_ctx, cert_file) != 1)
 	{
+	    *errstr = xasprintf(_("cannot load certificate file %s: %s"), 
+		    cert_file, ERR_error_string(ERR_get_error(), NULL));
 	    SSL_CTX_free(tls->ssl_ctx);
 	    tls->ssl_ctx = NULL;
-	    snprintf(errstr, errstr_bufsize,
-		    "cannot load certificate file %s: %s", 
-		    cert_file, ERR_error_string(ERR_get_error(), NULL));
 	    return TLS_EFILE;
 	}
     }
@@ -1012,22 +1032,20 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
     {
 	if (SSL_CTX_load_verify_locations(tls->ssl_ctx, trust_file, NULL) != 1)
 	{
+	    *errstr = xasprintf(_("cannot load trust file %s: %s"), 
+		    trust_file, ERR_error_string(ERR_get_error(), NULL));
 	    SSL_CTX_free(tls->ssl_ctx);
 	    tls->ssl_ctx = NULL;
-	    snprintf(errstr, errstr_bufsize,
-		    "cannot load CA file %s: %s", 
-		    trust_file, ERR_error_string(ERR_get_error(), NULL));
 	    return TLS_EFILE;
 	}
 	tls->have_trust_file = 1;
     }
     if (!(tls->ssl = SSL_new(tls->ssl_ctx)))
     {
+	*errstr = xasprintf(_("cannot create a TLS structure: %s"), 
+		ERR_error_string(ERR_get_error(), NULL));
 	SSL_CTX_free(tls->ssl_ctx);
 	tls->ssl_ctx = NULL;
-	snprintf(errstr, errstr_bufsize, 
-		"cannot create a TLS structure: %s", 
-		ERR_error_string(ERR_get_error(), NULL));
 	return TLS_ELIBFAILED;
     }
     return TLS_EOK;
@@ -1046,14 +1064,14 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
  * originates from an SSL_get_error() call with 'error_code' as its argument).
  * The line will read: "error_string: error_reason". 'error_string' is given by
  * the calling function, this function finds out 'error_reason'.
- * The resulting string will be returned in 'errstr'.
+ * The resulting string will be returned in an allocated string.
  * OpenSSL error strings are max 120 characters long according to
  * ERR_error_string(3). 
  */
 
 #ifdef HAVE_OPENSSL
-void openssl_io_error(int error_code, int error_code2, 
-	const char *error_string, char *errstr)
+char *openssl_io_error(int error_code, int error_code2, 
+	const char *error_string)
 {
     unsigned long error_code3;
     const char *error_reason;
@@ -1066,7 +1084,7 @@ void openssl_io_error(int error_code, int error_code2,
 	    {
 		if (error_code == 0)
 		{
-		    error_reason = "a protocol violating EOF occured";
+		    error_reason = _("a protocol violating EOF occured");
 		}
 		else if (error_code == -1)
 		{
@@ -1074,7 +1092,7 @@ void openssl_io_error(int error_code, int error_code2,
 		}
 		else
 		{
-		    error_reason = "unknown error";
+		    error_reason = _("unknown error");
 		}
 	    }
 	    else
@@ -1084,20 +1102,24 @@ void openssl_io_error(int error_code, int error_code2,
 	    break;
 	    
 	case SSL_ERROR_ZERO_RETURN:
-	    error_reason = "the connection has been closed unexpectedly";
+	    error_reason = _("the connection was closed unexpectedly");
 	    break;
 
 	case SSL_ERROR_SSL:
 	    error_reason = ERR_error_string(ERR_get_error(), NULL);
 	    break;
 	    
+	case SSL_ERROR_WANT_READ:
+	case SSL_ERROR_WANT_WRITE:
+    	    error_reason = _("the operation timed out");
+	    break;
+
 	default:
 	    /* probably SSL_ERROR_NONE */
-    	    error_reason = "unknown error";
+    	    error_reason = _("unknown error");
 	    break;
     }
-    snprintf(errstr, errstr_bufsize, 
-	    "%s: %s", error_string, error_reason);
+    return xasprintf("%s: %s", error_string, error_reason);
 }
 #endif /* HAVE_OPENSSL */
 
@@ -1109,7 +1131,7 @@ void openssl_io_error(int error_code, int error_code2,
  */
 
 int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck, 
-	tls_cert_info_t *tci, char *errstr)
+	tls_cert_info_t *tci, char **errstr)
 {
 #ifdef HAVE_GNUTLS
     int error_code;
@@ -1117,10 +1139,24 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
     gnutls_transport_set_ptr(tls->session, (gnutls_transport_ptr_t)fd);
     if ((error_code = gnutls_handshake(tls->session)) < 0)
     {
+	if (error_code == GNUTLS_E_INTERRUPTED)
+	{
+	    *errstr = xasprintf(_("operation aborted"));
+	}
+	else if (error_code == GNUTLS_E_AGAIN)
+	{
+	    /* This error message makes more sense than what
+	     * gnutls_strerror() would return. */
+	    *errstr = xasprintf(_("TLS handshake failed: %s"), 
+		    _("the operation timed out"));
+	}
+	else
+	{
+	    *errstr = xasprintf(_("TLS handshake failed: %s"), 
+	    	    gnutls_strerror(error_code));
+	}
 	gnutls_deinit(tls->session);
 	gnutls_certificate_free_credentials(tls->cred);
-	snprintf(errstr, errstr_bufsize, 
-		"TLS handshake failed: %s", gnutls_strerror(error_code));
 	return TLS_EHANDSHAKE;
     }
     if (tci)
@@ -1151,19 +1187,29 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
     
     if (!SSL_set_fd(tls->ssl, fd))
     {
+	*errstr = xasprintf(_("cannot set the file descriptor for TLS: %s"), 
+		ERR_error_string(ERR_get_error(), NULL));
 	SSL_free(tls->ssl);
 	SSL_CTX_free(tls->ssl_ctx);
-	snprintf(errstr, errstr_bufsize,
-		"cannot set the file descriptor for TLS: %s", 
-		ERR_error_string(ERR_get_error(), NULL));
 	return TLS_ELIBFAILED;
     }
     if ((error_code = SSL_connect(tls->ssl)) < 1)
     {
+	if (errno == EINTR 
+		&& (SSL_get_error(tls->ssl, error_code) == SSL_ERROR_WANT_READ
+		    || SSL_get_error(tls->ssl, error_code) 
+		    == SSL_ERROR_WANT_WRITE))
+	{
+	    *errstr = xasprintf(_("operation aborted"));
+	}
+	else
+	{
+	    *errstr = openssl_io_error(error_code, 
+		    SSL_get_error(tls->ssl, error_code), 
+		    _("TLS handshake failed"));
+	}
 	SSL_free(tls->ssl);
 	SSL_CTX_free(tls->ssl_ctx);
-	openssl_io_error(error_code, SSL_get_error(tls->ssl, error_code),
-		"TLS handshake failed", errstr);
 	return TLS_EIO;
     }
     if (tci)
@@ -1197,16 +1243,12 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
  * see tls.h
  */
 
-int tls_getchar(tls_t *tls, char *c, int *eof, char *errstr)
+int tls_getchar(tls_t *tls, char *c, int *eof, char **errstr)
 {
 #ifdef HAVE_GNUTLS
     ssize_t ret;
     
-    do
-    {
-	ret = gnutls_record_recv(tls->session, c, 1);
-    }
-    while (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);
+    ret = gnutls_record_recv(tls->session, c, 1);
     if (ret == 1)
     {
 	*eof = 0;
@@ -1219,9 +1261,22 @@ int tls_getchar(tls_t *tls, char *c, int *eof, char *errstr)
     }
     else
     {
-	snprintf(errstr, errstr_bufsize,
-		"cannot read from TLS connection: %s", 
-		gnutls_strerror(ret));
+	if (ret == GNUTLS_E_INTERRUPTED)
+	{
+	    *errstr = xasprintf(_("operation aborted"));
+	}
+	else if (ret == GNUTLS_E_AGAIN)
+	{
+	    /* This error message makes more sense than what
+	     * gnutls_strerror() would return. */
+	    *errstr = xasprintf(_("cannot read from TLS connection: %s"), 
+	    	    _("the operation timed out"));
+	}
+	else
+	{
+	    *errstr = xasprintf(_("cannot read from TLS connection: %s"), 
+	    	    gnutls_strerror(ret));
+	}
 	return TLS_EIO;
     }
     
@@ -1242,8 +1297,18 @@ int tls_getchar(tls_t *tls, char *c, int *eof, char *errstr)
 	}
 	else
 	{
-	    openssl_io_error(error_code, error_code2, 
-		    "cannot read from TLS connection", errstr);
+	    if (errno == EINTR &&
+	    	    (SSL_get_error(tls->ssl, error_code) == SSL_ERROR_WANT_READ
+		     || SSL_get_error(tls->ssl, error_code) 
+		     == SSL_ERROR_WANT_WRITE))
+	    {
+		*errstr = xasprintf(_("operation aborted"));
+	    }
+	    else
+    	    {
+    		*errstr = openssl_io_error(error_code, error_code2, 
+    			_("cannot read from TLS connection"));
+    	    }
 	    return TLS_EIO;
 	}
     }
@@ -1252,6 +1317,7 @@ int tls_getchar(tls_t *tls, char *c, int *eof, char *errstr)
 	*eof = 0;
 	return TLS_EOK;
     }
+    
 #endif /* HAVE_OPENSSL */
 }
 
@@ -1262,7 +1328,7 @@ int tls_getchar(tls_t *tls, char *c, int *eof, char *errstr)
  * see tls.h
  */
 
-int tls_gets(tls_t *tls, char *str, size_t size, size_t *len, char *errstr)
+int tls_gets(tls_t *tls, char *str, size_t size, size_t *len, char **errstr)
 {
     char c;
     size_t i;
@@ -1301,9 +1367,10 @@ int tls_gets(tls_t *tls, char *str, size_t size, size_t *len, char *errstr)
  * see tls.h
  */
 
-int tls_puts(tls_t *tls, const char *s, size_t len, char *errstr)
+int tls_puts(tls_t *tls, const char *s, size_t len, char **errstr)
 {
 #ifdef HAVE_GNUTLS
+    
     ssize_t ret;
 
     if (len < 1)
@@ -1312,16 +1379,24 @@ int tls_puts(tls_t *tls, const char *s, size_t len, char *errstr)
 	return TLS_EOK;
     }
     
-    do
+    if ((ret = gnutls_record_send(tls->session, s, len)) < 0)
     {
-	ret = gnutls_record_send(tls->session, s, len);
-    }
-    while (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);
-    if (ret < 0)
-    {
-	snprintf(errstr, errstr_bufsize,
-		"cannot write to TLS connection: %s", 
-		gnutls_strerror(ret));
+	if (ret == GNUTLS_E_INTERRUPTED)
+	{
+	    *errstr = xasprintf(_("operation aborted"));
+	}
+	else if (ret == GNUTLS_E_AGAIN)
+	{
+	    /* This error message makes more sense than what
+	     * gnutls_strerror() would return. */
+	    *errstr = xasprintf(_("cannot write to TLS connection: %s"), 
+		    _("the operation timed out"));
+	}
+	else
+	{
+	    *errstr = xasprintf(_("cannot write to TLS connection: %s"), 
+	    	    gnutls_strerror(ret));
+	}
 	return TLS_EIO;
     }
     else if ((size_t)ret == len)
@@ -1330,8 +1405,8 @@ int tls_puts(tls_t *tls, const char *s, size_t len, char *errstr)
     }
     else /* 0 <= error_code < len */
     {
-	snprintf(errstr, errstr_bufsize,
-		"cannot write to TLS connection: unknown error");
+	*errstr = xasprintf(_("cannot write to TLS connection: %s"), 
+		_("unknown error"));
 	return TLS_EIO;
     }    
 
@@ -1349,12 +1424,24 @@ int tls_puts(tls_t *tls, const char *s, size_t len, char *errstr)
     
     if ((error_code = SSL_write(tls->ssl, s, (int)len)) != (int)len)
     {
-	openssl_io_error(error_code, SSL_get_error(tls->ssl, error_code),
-		    "cannot write to TLS connection", errstr);
+	if (errno == EINTR 
+		&& ((SSL_get_error(tls->ssl, error_code) == SSL_ERROR_WANT_READ
+		       	|| SSL_get_error(tls->ssl, error_code) 
+			== SSL_ERROR_WANT_WRITE)))
+	{
+	    *errstr = xasprintf(_("operation aborted"));	    
+	}
+	else
+	{
+	    *errstr = openssl_io_error(error_code,
+	    	    SSL_get_error(tls->ssl, error_code),
+	    	    _("cannot write to TLS connection"));
+	}
 	return TLS_EIO;
     }
 
     return TLS_EOK;
+
 #endif /* HAVE_OPENSSL */
 }
 
@@ -1370,7 +1457,7 @@ void tls_close(tls_t *tls)
     if (tls->is_active)
     {
 #ifdef HAVE_GNUTLS
-	gnutls_bye(tls->session, GNUTLS_SHUT_RDWR);
+	gnutls_bye(tls->session, GNUTLS_SHUT_WR);
 	gnutls_deinit(tls->session);
 	gnutls_certificate_free_credentials(tls->cred);
 #endif /* HAVE_GNUTLS */
